% Fulvio Babich, University of Trieste, babich@units.it
% ============================================================================================
% Script containing some examples about Hamming decoding:
clear;
% SEZIONE DI INPUT
%ridondanza
R=8;
dim=4; %<=5
iterations=3; %numero massimo iterazioni
% intervallo di test
ebn0dB=[2.0:0.25:4.0]; 
%numero di simulazioni
max_test=1000000;
%condizone di arresto
max_err=100;
%intervallo visualizzazione
interval=20000;
%colore linee
main_color='k';
% FINE SEZIONE DI INPUT

str_theory=sprintf('--%s',main_color);
str_sim=sprintf(':*%s',main_color);
res=sprintf('R=%d, dim=%d, iterations=%d', R,dim,iterations); disp(res);

%INIZIALIZZAZIONE
%dimensione blocoo
N=(R+1)^dim;
%tasso del codice
Rc=(R/(R+1))^dim;
%distanza minima
dmin=2^dim;
%numero di parole a distanza minima
Ndmin=R^dim*(R+1)^dim/dmin;
switch dim
    case 2
        word=ones(R+1,R+1);
    case 3
        word=ones(R+1,R+1,R+1);
    case 4
        word=ones(R+1,R+1,R+1,R+1);
    case 5
        word=ones(R+1,R+1,R+1,R+1,R+1);
    otherwise
        disp('errore dim')
        return;
end
snrv=10.^(ebn0dB/10)*Rc;
% approssimazione probabilit d'errore soft di parola teorica
pwst=1/2*Ndmin*erfc((snrv*dmin).^.5);
% approssimazione probabilit d'errore soft di bit teorica
pbst=pwst*dmin/N;
%probabilit d'errore di bit senza codice
pbnc=1/2*erfc((10.^(ebn0dB/10)).^.5);
%inizializzazione contatori
snr_l=length(snrv);
errors_ws=zeros(1,snr_l);
errors_wh=zeros(1,snr_l);
errors_bs=zeros(1,snr_l);
test_done=zeros(1,snr_l);

%SIMULAZIONE
for il=1:snr_l
    snr=snrv(il);
    while test_done(il)<max_test & errors_ws(il)<max_err
        test_done(il)=test_done(il)+1;
        noise=randn(size(word))/(2*snr)^.5;
        rec=word+noise;
        llr=4*rec*snr^.5;
        if dim==2
            errors=sum(sum(llr<0));
        %hard decoding
            if errors > 1
                errors_wh(il)=errors_wh(il)+1;
            end
        end
if rem(test_done(il),interval)==0, res=sprintf('ebn0dB %.1f, test %6d, errors %3d', ebn0dB(il),test_done(il),errors_ws(il)); disp(res); end
%soft decoding
        llrs=llr;
        errors=1; ii=0;
        while errors & ii < iterations
            ii=ii+1;
            switch dim
                case 2
                    for i=1:R+1, temp=tanh(llrs/2); temp(i,:)=1; px(i,:)=2*atanh(prod(temp,1)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,i)=1; px(:,i)=2*atanh(prod(temp,2)); end, llrs=llrs+px;
                    errors=sum(sum(llrs<0));
                case 3
                    for i=1:R+1, temp=tanh(llrs/2); temp(i,:,:)=1; px(i,:,:)=2*atanh(prod(temp,1)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,i,:)=1; px(:,i,:)=2*atanh(prod(temp,2)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,:,i)=1; px(:,:,i)=2*atanh(prod(temp,3)); end, llrs=llrs+px;
                    errors=sum(sum(sum(llrs<0)));
                case 4
                    for i=1:R+1, temp=tanh(llrs/2); temp(i,:,:,:)=1; px(i,:,:,:)=2*atanh(prod(temp,1)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,i,:,:)=1; px(:,i,:,:)=2*atanh(prod(temp,2)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,:,i,:)=1; px(:,:,i,:)=2*atanh(prod(temp,3)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,:,:,i)=1; px(:,:,:,i)=2*atanh(prod(temp,4)); end, llrs=llrs+px;
                    errors=sum(sum(sum(sum(llrs<0))));
                 case 5
                    for i=1:R+1, temp=tanh(llrs/2); temp(i,:,:,:,:)=1; px(i,:,:,:,:)=2*atanh(prod(temp,1)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,i,:,:,:)=1; px(:,i,:,:,:)=2*atanh(prod(temp,2)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,:,i,:,:)=1; px(:,:,i,:,:)=2*atanh(prod(temp,3)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,:,:,i,:)=1; px(:,:,:,i,:)=2*atanh(prod(temp,4)); end, llrs=llrs+px;
                    for i=1:R+1, temp=tanh(llrs/2); temp(:,:,:,:,i)=1; px(:,:,:,:,i)=2*atanh(prod(temp,5)); end, llrs=llrs+px;
                    errors=sum(sum(sum(sum(sum(llrs<0)))));
           end
        end
        if errors > 0
            errors_ws(il)=errors_ws(il)+1;
            errors_bs(il)=errors_bs(il)+errors;
        end
    end
    res=sprintf('ebn0dB %.1f, test %6d, errors %3d, error rate:%.2e', ebn0dB(il),test_done(il),errors_ws(il),errors_ws(il)/test_done(il)); disp(res);
end
pws=errors_ws./test_done;
pbs=errors_bs./test_done/N;
phs=errors_wh./test_done;
figure(1)
semilogy(ebn0dB,pws,'b:*');
set(gca,'FontName','TimesNewRoman');
set(gca,'FontSize',12);
set(gca,'YLim',[1e-5,1]);
xlabel('E_b/N_0 [dB]');
ylabel('BLER');
hold on
semilogy(ebn0dB,pwst,str_theory);
if dim==2
    semilogy(ebn0dB,phs,str_sim);
end
figure(2)
semilogy(ebn0dB,pbs,str_sim);
set(gca,'FontName','TimesNewRoman');
set(gca,'FontSize',12);
set(gca,'YLim',[1e-6,1]);
xlabel('E_b/N_0 [dB]');
ylabel('BER');
hold on
semilogy(ebn0dB,pbst,str_theory);
semilogy(ebn0dB,pbnc,'k:');


